@@ -1,6 +1,6 @@
 /*
  * LibSylph -- E-Mail client library
- * Copyright (C) 1999-2007 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2011 Hiroyuki Yamamoto
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -54,12 +54,18 @@ typedef enum
 	FLT_O_REGEX	= 1 << 2
 } FilterOldFlag;
 
+static FilterInAddressBookFunc default_addrbook_func = NULL;
+
 static gboolean filter_match_cond	(FilterCond	*cond,
 					 MsgInfo	*msginfo,
 					 GSList		*hlist,
 					 FilterInfo	*fltinfo);
 static gboolean filter_match_header_cond(FilterCond	*cond,
 					 GSList		*hlist);
+static gboolean filter_match_in_addressbook
+					(FilterCond	*cond,
+					 GSList		*hlist,
+					 FilterInfo	*fltinfo);
 
 static void filter_cond_free		(FilterCond	*cond);
 static void filter_action_free		(FilterAction	*action);
@@ -79,6 +85,11 @@ gint filter_apply(GSList *fltlist, const gchar *file, FilterInfo *fltinfo)
 	if (!msginfo) return 0;
 	msginfo->file_path = g_strdup(file);
 
+	/* inherit MIME flag */
+	fltinfo->flags.tmp_flags =
+		(fltinfo->flags.tmp_flags & ~MSG_CACHED_FLAG_MASK) |
+		(msginfo->flags.tmp_flags & MSG_CACHED_FLAG_MASK);
+
 	ret = filter_apply_msginfo(fltlist, msginfo, fltinfo);
 
 	procmsg_msginfo_free(msginfo);
@@ -110,6 +121,8 @@ gint filter_apply_msginfo(GSList *fltlist, MsgInfo *msginfo,
 		return 0;
 	}
 
+	procmsg_set_auto_decrypt_message(FALSE);
+
 	for (cur = fltlist; cur != NULL; cur = cur->next) {
 		gboolean matched;
 
@@ -120,6 +133,8 @@ gint filter_apply_msginfo(GSList *fltlist, MsgInfo *msginfo,
 			g_warning("filter_match_rule() returned error (code: %d)\n", fltinfo->error);
 		}
 		if (matched) {
+			debug_print("filter-log: %s: rule [%s] matched\n",
+				    G_STRFUNC, rule->name ? rule->name : "(No name)");
 			ret = filter_action_exec(rule, msginfo, file, fltinfo);
 			if (ret < 0) {
 				g_warning("filter_action_exec() returned error (code: %d)\n", fltinfo->error);
@@ -131,6 +146,8 @@ gint filter_apply_msginfo(GSList *fltlist, MsgInfo *msginfo,
 		}
 	}
 
+	procmsg_set_auto_decrypt_message(TRUE);
+
 	procheader_header_list_destroy(hlist);
 	g_free(file);
 
@@ -145,6 +162,7 @@ gint filter_action_exec(FilterRule *rule, MsgInfo *msginfo, const gchar *file,
 	GSList *cur;
 	gchar *cmdline;
 	gboolean copy_to_self = FALSE;
+	gint ret;
 
 	g_return_val_if_fail(rule != NULL, -1);
 	g_return_val_if_fail(msginfo != NULL, -1);
@@ -183,14 +201,28 @@ gint filter_action_exec(FilterRule *rule, MsgInfo *msginfo, const gchar *file,
 		case FLT_ACTION_EXEC:
 			cmdline = g_strconcat(action->str_value, " \"", file,
 					      "\"", NULL);
-			execute_command_line(cmdline, FALSE);
+			ret = execute_command_line(cmdline, FALSE);
+			fltinfo->last_exec_exit_status = ret;
+			if (ret == -1) {
+				fltinfo->error = FLT_ERROR_EXEC_FAILED;
+				g_warning("filter_action_exec: cannot execute command: %s", cmdline);
+				g_free(cmdline);
+				return -1;
+			}
 			g_free(cmdline);
 			fltinfo->actions[action->type] = TRUE;
 			break;
 		case FLT_ACTION_EXEC_ASYNC:
 			cmdline = g_strconcat(action->str_value, " \"", file,
 					      "\"", NULL);
-			execute_command_line(cmdline, TRUE);
+			ret = execute_command_line(cmdline, TRUE);
+			fltinfo->last_exec_exit_status = ret;
+			if (ret == -1) {
+				fltinfo->error = FLT_ERROR_EXEC_FAILED;
+				g_warning("filter_action_exec: cannot execute command: %s", cmdline);
+				g_free(cmdline);
+				return -1;
+			}
 			g_free(cmdline);
 			fltinfo->actions[action->type] = TRUE;
 			break;
@@ -240,12 +272,17 @@ gint filter_action_exec(FilterRule *rule, MsgInfo *msginfo, const gchar *file,
 					fltinfo->actions[action->type] = TRUE;
 				}
 			} else {
-				if (folder_item_add_msg(dest_folder, file,
-							&fltinfo->flags,
-							FALSE) < 0) {
+				MsgFlags save_flags;
+
+				save_flags = msginfo->flags;
+				msginfo->flags = fltinfo->flags;
+				if (folder_item_add_msg_msginfo
+					(dest_folder, msginfo, FALSE) < 0) {
+					msginfo->flags = save_flags;
 					fltinfo->error = FLT_ERROR_ERROR;
 					return -1;
 				}
+				msginfo->flags = save_flags;
 				fltinfo->actions[action->type] = TRUE;
 			}
 
@@ -346,20 +383,66 @@ gboolean filter_match_rule(FilterRule *rule, MsgInfo *msginfo, GSList *hlist,
 	if (rule->bool_op == FLT_AND) {
 		for (cur = rule->cond_list; cur != NULL; cur = cur->next) {
 			cond = (FilterCond *)cur->data;
-			matched = filter_match_cond(cond, msginfo, hlist,
-						    fltinfo);
-			if (matched == FALSE)
-				return FALSE;
+			if (cond->type >= FLT_COND_SIZE_GREATER) {
+				matched = filter_match_cond
+					(cond, msginfo, hlist, fltinfo);
+				if (matched == FALSE)
+					return FALSE;
+			}
+		}
+
+		for (cur = rule->cond_list; cur != NULL; cur = cur->next) {
+			cond = (FilterCond *)cur->data;
+			if (cond->type <= FLT_COND_TO_OR_CC) {
+				matched = filter_match_cond
+					(cond, msginfo, hlist, fltinfo);
+				if (matched == FALSE)
+					return FALSE;
+			}
+		}
+
+		for (cur = rule->cond_list; cur != NULL; cur = cur->next) {
+			cond = (FilterCond *)cur->data;
+			if (cond->type == FLT_COND_BODY ||
+			    cond->type == FLT_COND_CMD_TEST) {
+				matched = filter_match_cond
+					(cond, msginfo, hlist, fltinfo);
+				if (matched == FALSE)
+					return FALSE;
+			}
 		}
 
 		return TRUE;
 	} else if (rule->bool_op == FLT_OR) {
 		for (cur = rule->cond_list; cur != NULL; cur = cur->next) {
 			cond = (FilterCond *)cur->data;
-			matched = filter_match_cond(cond, msginfo, hlist,
-						    fltinfo);
-			if (matched == TRUE)
-				return TRUE;
+			if (cond->type >= FLT_COND_SIZE_GREATER) {
+				matched = filter_match_cond
+					(cond, msginfo, hlist, fltinfo);
+				if (matched == TRUE)
+					return TRUE;
+			}
+		}
+
+		for (cur = rule->cond_list; cur != NULL; cur = cur->next) {
+			cond = (FilterCond *)cur->data;
+			if (cond->type <= FLT_COND_TO_OR_CC) {
+				matched = filter_match_cond
+					(cond, msginfo, hlist, fltinfo);
+				if (matched == TRUE)
+					return TRUE;
+			}
+		}
+
+		for (cur = rule->cond_list; cur != NULL; cur = cur->next) {
+			cond = (FilterCond *)cur->data;
+			if (cond->type == FLT_COND_BODY ||
+			    cond->type == FLT_COND_CMD_TEST) {
+				matched = filter_match_cond
+					(cond, msginfo, hlist, fltinfo);
+				if (matched == TRUE)
+					return TRUE;
+			}
 		}
 
 		return FALSE;
@@ -373,15 +456,24 @@ static gboolean filter_match_cond(FilterCond *cond, MsgInfo *msginfo,
 {
 	gint ret;
 	gboolean matched = FALSE;
+	gboolean not_match = FALSE;
 	gchar *file;
 	gchar *cmdline;
 	PrefsAccount *cond_ac;
 
 	switch (cond->type) {
 	case FLT_COND_HEADER:
+		if (cond->match_type == FLT_IN_ADDRESSBOOK)
+			return filter_match_in_addressbook(cond, hlist, fltinfo);
+		else
+			return filter_match_header_cond(cond, hlist);
 	case FLT_COND_ANY_HEADER:
-	case FLT_COND_TO_OR_CC:
 		return filter_match_header_cond(cond, hlist);
+	case FLT_COND_TO_OR_CC:
+		if (cond->match_type == FLT_IN_ADDRESSBOOK)
+			return filter_match_in_addressbook(cond, hlist, fltinfo);
+		else
+			return filter_match_header_cond(cond, hlist);
 	case FLT_COND_BODY:
 		matched = procmime_find_string(msginfo, cond->str_value,
 					       cond->match_func);
@@ -391,7 +483,8 @@ static gboolean filter_match_cond(FilterCond *cond, MsgInfo *msginfo,
 		if (!file)
 			return FALSE;
 		cmdline = g_strconcat(cond->str_value, " \"", file, "\"", NULL);
-		ret = execute_command_line(cmdline, FALSE);
+		ret = execute_command_line_async_wait(cmdline);
+		fltinfo->last_exec_exit_status = ret;
 		matched = (ret == 0);
 		if (ret == -1)
 			fltinfo->error = FLT_ERROR_EXEC_FAILED;
@@ -428,8 +521,47 @@ static gboolean filter_match_cond(FilterCond *cond, MsgInfo *msginfo,
 		return FALSE;
 	}
 
-	if (FLT_IS_NOT_MATCH(cond->match_flag))
+	if (FLT_IS_NOT_MATCH(cond->match_flag)) {
+		not_match = TRUE;
 		matched = !matched;
+	}
+
+	if (matched && get_debug_mode()) {
+		gchar *sv = cond->str_value ? cond->str_value : "";
+		gchar *nm = not_match ? " (reverse match)" : "";
+
+		switch (cond->type) {
+		case FLT_COND_BODY:
+			debug_print("filter-log: %s: BODY, str_value: [%s]%s\n", G_STRFUNC, sv, nm);
+			break;
+		case FLT_COND_CMD_TEST:
+			debug_print("filter-log: %s: CMD_TEST, str_value: [%s]%s\n", G_STRFUNC, sv, nm);
+			break;
+		case FLT_COND_SIZE_GREATER:
+			debug_print("filter-log: %s: SIZE_GREATER: %u %s %d (KB)%s\n", G_STRFUNC, msginfo->size, not_match ? "<=" : ">", cond->int_value, nm);
+			break;
+		case FLT_COND_AGE_GREATER:
+			debug_print("filter-log: %s: AGE_GREATER: %ld (sec) %s %d (day)%s\n", G_STRFUNC, time(NULL) - msginfo->date_t, not_match ? "<=" : ">", cond->int_value, nm);
+			break;
+		case FLT_COND_UNREAD:
+			debug_print("filter-log: %s: UNREAD%s\n", G_STRFUNC, nm);
+			break;
+		case FLT_COND_MARK:
+			debug_print("filter-log: %s: MARK%s\n", G_STRFUNC, nm);
+			break;
+		case FLT_COND_COLOR_LABEL:
+			debug_print("filter-log: %s: COLOR_LABEL%s\n", G_STRFUNC, nm);
+			break;
+		case FLT_COND_MIME:
+			debug_print("filter-log: %s: MIME%s\n", G_STRFUNC, nm);
+			break;
+		case FLT_COND_ACCOUNT:
+			debug_print("filter-log: %s: ACCOUNT [%d]%s\n", G_STRFUNC, cond->int_value, nm);
+			break;
+		default:
+			break;
+		}
+	}
 
 	return matched;
 }
@@ -437,6 +569,7 @@ static gboolean filter_match_cond(FilterCond *cond, MsgInfo *msginfo,
 static gboolean filter_match_header_cond(FilterCond *cond, GSList *hlist)
 {
 	gboolean matched = FALSE;
+	gboolean not_match = FALSE;
 	GSList *cur;
 	Header *header;
 
@@ -475,8 +608,86 @@ static gboolean filter_match_header_cond(FilterCond *cond, GSList *hlist)
 			break;
 	}
 
-	if (FLT_IS_NOT_MATCH(cond->match_flag))
+	if (FLT_IS_NOT_MATCH(cond->match_flag)) {
+		not_match = TRUE;
 		matched = !matched;
+	}
+
+	if (matched && get_debug_mode()) {
+		gchar *sv = cond->str_value ? cond->str_value : "";
+		gchar *nm = not_match ? " (reverse match)" : "";
+
+		switch (cond->type) {
+		case FLT_COND_HEADER:
+			debug_print("filter-log: %s: HEADER [%s], str_value: [%s]%s\n", G_STRFUNC, cond->header_name, sv, nm);
+			break;
+		case FLT_COND_ANY_HEADER:
+			debug_print("filter-log: %s: ANY_HEADER, str_value: [%s]%s\n", G_STRFUNC, sv, nm);
+			break;
+		case FLT_COND_TO_OR_CC:
+			debug_print("filter-log: %s: TO_OR_CC, str_value: [%s]%s\n", G_STRFUNC, sv, nm);
+			break;
+		default:
+			break;
+		}
+	}
+
+	return matched;
+}
+
+static gboolean filter_match_in_addressbook(FilterCond *cond, GSList *hlist,
+					    FilterInfo *fltinfo)
+{
+	gboolean matched = FALSE;
+	gboolean not_match = FALSE;
+	GSList *cur;
+	Header *header;
+
+	if (!default_addrbook_func)
+		return FALSE;
+	if (cond->type != FLT_COND_HEADER && cond->type != FLT_COND_TO_OR_CC)
+		return FALSE;
+
+	for (cur = hlist; cur != NULL; cur = cur->next) {
+		header = (Header *)cur->data;
+
+		if (cond->type == FLT_COND_HEADER) {
+			if (!g_ascii_strcasecmp
+				(header->name, cond->header_name)) {
+				if (default_addrbook_func(header->body))
+					matched = TRUE;
+			}
+		} else if (cond->type == FLT_COND_TO_OR_CC) {
+			if (!g_ascii_strcasecmp(header->name, "To") ||
+			    !g_ascii_strcasecmp(header->name, "Cc")) {
+				if (default_addrbook_func(header->body))
+					matched = TRUE;
+			}
+		}
+
+		if (matched == TRUE)
+			break;
+	}
+
+	if (FLT_IS_NOT_MATCH(cond->match_flag)) {
+		not_match = TRUE;
+		matched = !matched;
+	}
+
+	if (matched && get_debug_mode()) {
+		gchar *nm = not_match ? " (reverse match)" : "";
+
+		switch (cond->type) {
+		case FLT_COND_HEADER:
+			debug_print("filter-log: %s: HEADER [%s], IN_ADDRESSBOOK%s\n", G_STRFUNC, cond->header_name, nm);
+			break;
+		case FLT_COND_TO_OR_CC:
+			debug_print("filter-log: %s: TO_OR_CC, IN_ADDRESSBOOK%s\n", G_STRFUNC, nm);
+			break;
+		default:
+			break;
+		}
+	}
 
 	return matched;
 }
@@ -812,7 +1023,8 @@ void filter_write_file(GSList *list, const gchar *file)
 		FilterRule *rule = (FilterRule *)cur->data;
 		GSList *cur_cond;
 		GSList *cur_action;
-		gchar match_type[16];
+		gchar match_type[64];
+		gchar nstr[16];
 
 		fputs("    <rule name=\"", pfile->fp);
 		xml_file_put_escape_str(pfile->fp, rule->name);
@@ -832,19 +1044,27 @@ void filter_write_file(GSList *list, const gchar *file)
 
 			switch (cond->match_type) {
 			case FLT_CONTAIN:
-				strcpy(match_type,
-				       FLT_IS_NOT_MATCH(cond->match_flag)
-				       ? "not-contain" : "contains");
+				strncpy2(match_type,
+					 FLT_IS_NOT_MATCH(cond->match_flag)
+					 ? "not-contain" : "contains",
+					 sizeof(match_type));
 				break;
 			case FLT_EQUAL:
-				strcpy(match_type,
-				       FLT_IS_NOT_MATCH(cond->match_flag)
-				       ? "is-not" : "is");
+				strncpy2(match_type,
+					 FLT_IS_NOT_MATCH(cond->match_flag)
+					 ? "is-not" : "is", sizeof(match_type));
 				break;
 			case FLT_REGEX:
-				strcpy(match_type,
-				       FLT_IS_NOT_MATCH(cond->match_flag)
-				       ? "not-regex" : "regex");
+				strncpy2(match_type,
+					 FLT_IS_NOT_MATCH(cond->match_flag)
+					 ? "not-regex" : "regex",
+					 sizeof(match_type));
+				break;
+			case FLT_IN_ADDRESSBOOK:
+				strncpy2(match_type,
+					 FLT_IS_NOT_MATCH(cond->match_flag)
+					 ? "not-in-addressbook" : "in-addressbook",
+					 sizeof(match_type));
 				break;
 			default:
 				match_type[0] = '\0';
@@ -881,13 +1101,13 @@ void filter_write_file(GSList *list, const gchar *file)
 				NODE_NEW("command-test", cond->str_value);
 				break;
 			case FLT_COND_SIZE_GREATER:
-				NODE_NEW("size", itos(cond->int_value));
+				NODE_NEW("size", itos_buf(nstr, cond->int_value));
 				ADD_ATTR("type",
 					 FLT_IS_NOT_MATCH(cond->match_flag)
 					 ? "lt" : "gt");
 				break;
 			case FLT_COND_AGE_GREATER:
-				NODE_NEW("age", itos(cond->int_value));
+				NODE_NEW("age", itos_buf(nstr, cond->int_value));
 				ADD_ATTR("type",
 					 FLT_IS_NOT_MATCH(cond->match_flag)
 					 ? "lt" : "gt");
@@ -917,7 +1137,7 @@ void filter_write_file(GSList *list, const gchar *file)
 					 ? "is-not" : "is");
 				break;
 			case FLT_COND_ACCOUNT:
-				 NODE_NEW("account-id", itos(cond->int_value));
+				 NODE_NEW("account-id", itos_buf(nstr, cond->int_value));
 				 break;
 			default:
 				 break;
@@ -1196,6 +1416,16 @@ FilterRule *filter_read_str(const gchar *str)
 	return rule;
 }
 
+void filter_set_addressbook_func(FilterInAddressBookFunc func)
+{
+	default_addrbook_func = func;
+}
+
+FilterInAddressBookFunc filter_get_addressbook_func(void)
+{
+	return default_addrbook_func;
+}
+
 FilterRule *filter_rule_new(const gchar *name, FilterBoolOp bool_op,
 			    GSList *cond_list, GSList *action_list)
 {
@@ -1231,7 +1461,8 @@ FilterCond *filter_cond_new(FilterCondType type,
 		cond->header_name = NULL;
 
 	cond->str_value = (value && *value) ? g_strdup(value) : NULL;
-	if (type == FLT_COND_SIZE_GREATER || type == FLT_COND_AGE_GREATER)
+	if (type == FLT_COND_SIZE_GREATER || type == FLT_COND_AGE_GREATER ||
+	    type == FLT_COND_ACCOUNT)
 		cond->int_value = atoi(value);
 	else
 		cond->int_value = 0;
@@ -1243,6 +1474,8 @@ FilterCond *filter_cond_new(FilterCondType type,
 			cond->match_func = str_find_equal;
 		else
 			cond->match_func = str_case_find_equal;
+	} else if (match_type == FLT_IN_ADDRESSBOOK) {
+		cond->match_func = str_case_find_equal;
 	} else {
 		if (FLT_IS_CASE_SENS(match_flag))
 			cond->match_func = str_find;
@@ -1278,10 +1511,111 @@ FilterInfo *filter_info_new(void)
 	fltinfo->move_dest = NULL;
 	fltinfo->drop_done = FALSE;
 	fltinfo->error = FLT_ERROR_OK;
+	fltinfo->last_exec_exit_status = 0;
 
 	return fltinfo;
 }
 
+FilterRule *filter_junk_rule_create(PrefsAccount *account,
+				    FolderItem *default_junk,
+				    gboolean is_manual)
+{
+	FilterRule *rule;
+	FilterCond *cond;
+	FilterAction *action;
+	GSList *cond_list = NULL, *action_list = NULL;
+	gchar *junk_id = NULL;
+	FolderItem *item = NULL;
+
+	if (!prefs_common.junk_classify_cmd)
+		return NULL;
+
+	if (prefs_common.junk_folder)
+		item = folder_find_item_from_identifier(prefs_common.junk_folder);
+
+	if (!item && account) {
+		Folder *folder = NULL;
+		GList *list;
+
+		/* find most suitable Junk folder for account */
+
+		if (account->inbox && *account->inbox == '#') {
+			FolderItem *inbox;
+			inbox = folder_find_item_from_identifier(account->inbox);
+			if (inbox) {
+				folder = inbox->folder;
+				if (folder)
+					item = folder_get_junk(folder);
+			}
+		}
+		if (!item) {
+			folder = FOLDER(account->folder);
+			if (folder)
+				item = folder_get_junk(folder);
+		}
+		if (!item) {
+			for (list = folder_get_list(); list != NULL; list = list->next) {
+				folder = FOLDER(list->data);
+				if (FOLDER_IS_LOCAL(folder)) {
+					if (folder->account == account)
+						item = folder_get_junk(folder);
+					if (!item && folder->node) {
+						item = FOLDER_ITEM(folder->node->data);
+						if (item && item->account == account && item->folder)
+							item = folder_get_junk(item->folder);
+						else
+							item = NULL;
+					}
+				}
+				if (item)
+					break;
+			}
+		}
+	}
+
+	if (!item)
+		item = default_junk;
+	if (!item)
+		item = folder_get_default_junk();
+	if (!item)
+		return NULL;
+	junk_id = folder_item_get_identifier(item);
+	if (!junk_id)
+		return NULL;
+
+	debug_print("filter_junk_rule_create: junk folder: %s\n",
+		    junk_id);
+
+	cond = filter_cond_new(FLT_COND_CMD_TEST, 0, 0, NULL,
+			       prefs_common.junk_classify_cmd);
+	cond_list = g_slist_append(NULL, cond);
+	if (prefs_common.delete_junk_on_recv && !is_manual) {
+		action = filter_action_new(FLT_ACTION_COPY, junk_id);
+		action_list = g_slist_append(NULL, action);
+		action = filter_action_new(FLT_ACTION_DELETE, NULL);
+		action_list = g_slist_append(action_list, action);
+	} else {
+		action = filter_action_new(FLT_ACTION_MOVE, junk_id);
+		action_list = g_slist_append(NULL, action);
+	}
+
+	if (prefs_common.mark_junk_as_read) {
+		action = filter_action_new(FLT_ACTION_MARK_READ, NULL);
+		action_list = g_slist_append(action_list, action);
+	}
+
+	if (is_manual)
+		rule = filter_rule_new(_("Junk mail filter (manual)"), FLT_OR,
+				       cond_list, action_list);
+	else
+		rule = filter_rule_new(_("Junk mail filter"), FLT_OR,
+				       cond_list, action_list);
+
+	g_free(junk_id);
+
+        return rule;
+}
+
 void filter_rule_rename_dest_path(FilterRule *rule, const gchar *old_path,
 				  const gchar *new_path)
 {
@@ -1411,6 +1745,11 @@ void filter_rule_match_type_str_to_enum(const gchar *match_type,
 	} else if (!strcmp(match_type, "not-regex")) {
 		*type = FLT_REGEX;
 		*flag = FLT_NOT_MATCH;
+	} else if (!strcmp(match_type, "in-addressbook")) {
+		*type = FLT_IN_ADDRESSBOOK;
+	} else if (!strcmp(match_type, "not-in-addressbook")) {
+		*type = FLT_IN_ADDRESSBOOK;
+		*flag = FLT_NOT_MATCH;
 	} else if (!strcmp(match_type, "gt")) {
 	} else if (!strcmp(match_type, "lt")) {
 		*flag = FLT_NOT_MATCH;
