@@ -1,6 +1,6 @@
 /*
  * LibSylph -- E-Mail client library
- * Copyright (C) 1999-2009 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2012 Hiroyuki Yamamoto
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -43,11 +43,26 @@
 #include "prefs_account.h"
 #include "sylmain.h"
 
+typedef struct _FolderPrivData FolderPrivData;
+
+struct _FolderPrivData {
+	Folder *folder;
+	FolderItem *junk;
+
+	FolderUIFunc2 ui_func2;
+	gpointer ui_func2_data;
+
+	gpointer data;
+};
+
 static GList *folder_list = NULL;
+static GList *folder_priv_list = NULL;
 
 static void folder_init		(Folder		*folder,
 				 const gchar	*name);
 
+static FolderPrivData *folder_get_priv	(Folder		*folder);
+
 static gboolean folder_read_folder_func	(GNode		*node,
 					 gpointer	 data);
 static gchar *folder_get_list_path	(void);
@@ -115,14 +130,23 @@ void folder_remote_folder_init(Folder *folder, const gchar *name,
 
 void folder_destroy(Folder *folder)
 {
+	FolderPrivData *priv;
+
 	g_return_if_fail(folder != NULL);
 	g_return_if_fail(folder->klass->destroy != NULL);
 
+	debug_print("folder_destroy: destroying Folder (%p)\n", folder);
+
 	folder->klass->destroy(folder);
 
 	folder_list = g_list_remove(folder_list, folder);
 
 	folder_tree_destroy(folder);
+
+	priv = folder_get_priv(folder);
+	folder_priv_list = g_list_remove(folder_priv_list, priv);
+	g_free(priv);
+
 	g_free(folder->name);
 	g_free(folder);
 }
@@ -142,6 +166,55 @@ void folder_remote_folder_destroy(RemoteFolder *rfolder)
 		session_destroy(rfolder->session);
 }
 
+gint folder_remote_folder_destroy_all_sessions(void)
+{
+	GList *list;
+	Folder *folder;
+	RemoteFolder *rfolder;
+
+	for (list = folder_list; list != NULL; list = list->next) {
+		folder = FOLDER(list->data);
+		if (FOLDER_IS_REMOTE(folder)) {
+			rfolder = REMOTE_FOLDER(folder);
+			if (rfolder->session &&
+			    !folder_remote_folder_is_session_active(rfolder)) {
+				session_destroy(rfolder->session);
+				rfolder->session = NULL;
+			}
+		}
+	}
+
+	return 0;
+}
+
+gboolean folder_remote_folder_is_session_active(RemoteFolder *rfolder)
+{
+	g_return_val_if_fail(rfolder != NULL, FALSE);
+
+	if (FOLDER_TYPE(rfolder) == F_IMAP)
+		return imap_is_session_active(IMAP_FOLDER(rfolder));
+
+	return FALSE;
+}
+
+gboolean folder_remote_folder_active_session_exist(void)
+{
+	GList *list;
+	Folder *folder;
+	RemoteFolder *rfolder;
+
+	for (list = folder_list; list != NULL; list = list->next) {
+		folder = FOLDER(list->data);
+		if (FOLDER_IS_REMOTE(folder)) {
+			rfolder = REMOTE_FOLDER(folder);
+			if (folder_remote_folder_is_session_active(rfolder))
+				return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
 gint folder_scan_tree(Folder *folder)
 {
 	g_return_val_if_fail(folder != NULL, -1);
@@ -312,6 +385,8 @@ void folder_item_destroy(FolderItem *item)
 			folder->queue = NULL;
 		else if (folder->trash == item)
 			folder->trash = NULL;
+		else if (folder_get_junk(folder) == item)
+			folder_set_junk(folder, NULL);
 	}
 
 	g_free(item->name);
@@ -371,6 +446,42 @@ void folder_set_ui_func(Folder *folder, FolderUIFunc func, gpointer data)
 	folder->ui_func_data = data;
 }
 
+void folder_set_ui_func2(Folder *folder, FolderUIFunc2 func, gpointer data)
+{
+	FolderPrivData *priv;
+
+	priv = folder_get_priv(folder);
+	if (priv) {
+		priv->ui_func2 = func;
+		priv->ui_func2_data = data;
+	}
+}
+
+FolderUIFunc2 folder_get_ui_func2(Folder *folder)
+{
+	FolderPrivData *priv;
+
+	priv = folder_get_priv(folder);
+	if (priv)
+		return priv->ui_func2;
+
+	return NULL;
+}
+
+gboolean folder_call_ui_func2(Folder *folder, FolderItem *item, guint count,
+			      guint total)
+{
+	FolderPrivData *priv;
+
+	priv = folder_get_priv(folder);
+	if (priv && priv->ui_func2) {
+		return priv->ui_func2(folder, item, count, total,
+				      priv->ui_func2_data);
+	}
+
+	return TRUE;
+}
+
 void folder_set_name(Folder *folder, const gchar *name)
 {
 	g_return_if_fail(folder != NULL);
@@ -398,6 +509,9 @@ void folder_add(Folder *folder)
 	Folder *cur_folder;
 	GList *cur;
 	gint i;
+	FolderPrivData *priv;
+
+	debug_print("Adding Folder (%p) to folder list\n", folder);
 
 	g_return_if_fail(folder != NULL);
 
@@ -416,6 +530,9 @@ void folder_add(Folder *folder)
 	}
 
 	folder_list = g_list_insert(folder_list, folder, i);
+	priv = g_new0(FolderPrivData, 1);
+	priv->folder = folder;
+	folder_priv_list = g_list_insert(folder_priv_list, priv, i);
 }
 
 GList *folder_get_list(void)
@@ -858,6 +975,77 @@ FolderItem *folder_get_default_trash(void)
 	return folder->trash;
 }
 
+FolderItem *folder_get_default_junk(void)
+{
+	FolderPrivData *priv;
+
+	if (!folder_list) return NULL;
+	if (!folder_priv_list) return NULL;
+	priv = (FolderPrivData *)folder_priv_list->data;
+	g_return_val_if_fail(priv != NULL, NULL);
+	g_return_val_if_fail(priv->folder != NULL, NULL);
+	return priv->junk;
+}
+
+static FolderPrivData *folder_get_priv(Folder *folder)
+{
+	FolderPrivData *priv;
+	GList *cur;
+
+	g_return_val_if_fail(folder != NULL, NULL);
+
+	for (cur = folder_priv_list; cur != NULL; cur = cur->next) {
+		priv = (FolderPrivData *)cur->data;
+		if (priv->folder == folder)
+			return priv;
+	}
+
+	g_warning("folder_get_priv: private data for Folder (%p) not found.",
+		  folder);
+
+	return NULL;
+}
+
+FolderItem *folder_get_junk(Folder *folder)
+{
+	FolderPrivData *priv;
+
+	priv = folder_get_priv(folder);
+	if (priv)
+		return priv->junk;
+
+	return NULL;
+}
+
+void folder_set_junk(Folder *folder, FolderItem *item)
+{
+	FolderPrivData *priv;
+
+	priv = folder_get_priv(folder);
+	if (priv)
+		priv->junk = item;
+}
+
+gboolean folder_item_is_trash(FolderItem *item)
+{
+	PrefsAccount *ac;
+	FolderItem *trash;
+
+	g_return_val_if_fail(item != NULL, FALSE);
+
+	if (item->stype == F_TRASH)
+		return TRUE;
+
+	ac = account_find_from_item_property(item);
+	if (ac && ac->set_trash_folder && ac->trash_folder) {
+		trash = folder_find_item_from_identifier(ac->trash_folder);
+		if (trash == item)
+			return TRUE;
+	}
+
+	return FALSE;
+}
+
 #define CREATE_FOLDER_IF_NOT_EXIST(member, dir, type)	\
 {							\
 	if (!folder->member) {				\
@@ -882,7 +1070,7 @@ void folder_set_missing_folders(void)
 		g_return_if_fail(rootitem != NULL);
 
 		if (folder->inbox && folder->outbox && folder->draft &&
-		    folder->queue && folder->trash)
+		    folder->queue && folder->trash && folder_get_junk(folder))
 			continue;
 
 		if (folder_create_tree(folder) < 0) {
@@ -896,7 +1084,13 @@ void folder_set_missing_folders(void)
 		CREATE_FOLDER_IF_NOT_EXIST(draft,  DRAFT_DIR,  F_DRAFT);
 		CREATE_FOLDER_IF_NOT_EXIST(queue,  QUEUE_DIR,  F_QUEUE);
 		CREATE_FOLDER_IF_NOT_EXIST(trash,  TRASH_DIR,  F_TRASH);
-		/* CREATE_FOLDER_IF_NOT_EXIST(junk,   JUNK_DIR,  F_JUNK); */
+
+		if (!folder_get_junk(folder)) {
+			item = folder_item_new(JUNK_DIR, JUNK_DIR);
+			item->stype = F_JUNK;
+			folder_item_append(rootitem, item);
+			folder_set_junk(folder, item);
+		}
 	}
 }
 
@@ -1185,6 +1379,10 @@ gint folder_item_add_msgs_msginfo(FolderItem *dest, GSList *msglist,
 					       remove_source, first);
 }
 
+#define IS_FROM_QUEUE(m, d) \
+	(m->folder && m->folder->stype == F_QUEUE && \
+	 MSG_IS_QUEUED(m->flags) && d->stype != F_QUEUE)
+
 gint folder_item_move_msg(FolderItem *dest, MsgInfo *msginfo)
 {
 	Folder *folder;
@@ -1195,12 +1393,21 @@ gint folder_item_move_msg(FolderItem *dest, MsgInfo *msginfo)
 
 	folder = dest->folder;
 
+	if (IS_FROM_QUEUE(msginfo, dest)) {
+		GSList msglist;
+
+		msglist.data = msginfo;
+		msglist.next = NULL;
+		return procmsg_add_messages_from_queue(dest, &msglist, TRUE);
+	}
+
 	return folder->klass->move_msg(folder, dest, msginfo);
 }
 
 gint folder_item_move_msgs(FolderItem *dest, GSList *msglist)
 {
 	Folder *folder;
+	MsgInfo *msginfo;
 
 	g_return_val_if_fail(dest != NULL, -1);
 	g_return_val_if_fail(msglist != NULL, -1);
@@ -1208,6 +1415,10 @@ gint folder_item_move_msgs(FolderItem *dest, GSList *msglist)
 
 	folder = dest->folder;
 
+	msginfo = (MsgInfo *)msglist->data;
+	if (IS_FROM_QUEUE(msginfo, dest))
+		return procmsg_add_messages_from_queue(dest, msglist, TRUE);
+
 	return folder->klass->move_msgs(folder, dest, msglist);
 }
 
@@ -1221,12 +1432,21 @@ gint folder_item_copy_msg(FolderItem *dest, MsgInfo *msginfo)
 
 	folder = dest->folder;
 
+	if (IS_FROM_QUEUE(msginfo, dest)) {
+		GSList msglist;
+
+		msglist.data = msginfo;
+		msglist.next = NULL;
+		return procmsg_add_messages_from_queue(dest, &msglist, FALSE);
+	}
+
 	return folder->klass->copy_msg(folder, dest, msginfo);
 }
 
 gint folder_item_copy_msgs(FolderItem *dest, GSList *msglist)
 {
 	Folder *folder;
+	MsgInfo *msginfo;
 
 	g_return_val_if_fail(dest != NULL, -1);
 	g_return_val_if_fail(msglist != NULL, -1);
@@ -1234,9 +1454,15 @@ gint folder_item_copy_msgs(FolderItem *dest, GSList *msglist)
 
 	folder = dest->folder;
 
+	msginfo = (MsgInfo *)msglist->data;
+	if (IS_FROM_QUEUE(msginfo, dest))
+		return procmsg_add_messages_from_queue(dest, msglist, FALSE);
+
 	return folder->klass->copy_msgs(folder, dest, msglist);
 }
 
+#undef IS_FROM_QUEUE
+
 gint folder_item_remove_msg(FolderItem *item, MsgInfo *msginfo)
 {
 	Folder *folder;
@@ -1392,10 +1618,8 @@ static gboolean folder_build_tree(GNode *node, gpointer data)
 				stype = F_QUEUE;
 			else if (!g_ascii_strcasecmp(attr->value, "trash"))
 				stype = F_TRASH;
-#if 0
 			else if (!g_ascii_strcasecmp(attr->value, "junk"))
 				stype = F_JUNK;
-#endif
 			else if (!g_ascii_strcasecmp(attr->value, "virtual"))
 				stype = F_VIRTUAL;
 		} else if (!strcmp(attr->name, "name"))
@@ -1519,7 +1743,7 @@ static gboolean folder_build_tree(GNode *node, gpointer data)
 	case F_DRAFT:  folder->draft  = item; break;
 	case F_QUEUE:  folder->queue  = item; break;
 	case F_TRASH:  folder->trash  = item; break;
-	/* case F_JUNK:   folder->junk   = item; break; */
+	case F_JUNK:   folder_set_junk(folder, item); break;
 	default:       break;
 	}
 	item->account = account;
