@@ -1,6 +1,6 @@
 /*
  * LibSylph -- E-Mail client library
- * Copyright (C) 1999-2010 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2012 Hiroyuki Yamamoto
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -47,6 +47,10 @@ static PrefParam param[] = {
 	 P_BOOL},
 	{"newmsg_notify_command", NULL, &prefs_common.newmsg_notify_cmd,
 	 P_STRING},
+	{"enable_newmsg_notify_sound", "FALSE",
+	 &prefs_common.enable_newmsg_notify_sound, P_BOOL},
+	{"newmsg_notify_sound", NULL, &prefs_common.newmsg_notify_sound,
+	 P_STRING},
 
 	{"inc_local", "FALSE", &prefs_common.inc_local, P_BOOL},
 	{"filter_on_inc_local", "TRUE", &prefs_common.filter_on_inc, P_BOOL},
@@ -305,6 +309,10 @@ static PrefParam param[] = {
 	{"addressbook_y", "32", &prefs_common.addressbook_y, P_INT},
 	{"addressbook_width", "620", &prefs_common.addressbook_width, P_INT},
 	{"addressbook_height", "360", &prefs_common.addressbook_height, P_INT},
+	{"addressbook_folder_width", "190", &prefs_common.addressbook_folder_width, P_INT},
+	{"addressbook_col_name", "164", &prefs_common.addressbook_col_name, P_INT},
+	{"addressbook_col_addr", "156", &prefs_common.addressbook_col_addr, P_INT},
+	{"addressbook_col_rem", "100", &prefs_common.addressbook_col_rem, P_INT},
 
 	/* Message */
 	{"enable_color", "TRUE", &prefs_common.enable_color, P_BOOL},
@@ -320,6 +328,11 @@ static PrefParam param[] = {
 	{"convert_mb_alnum", "FALSE", &prefs_common.conv_mb_alnum, P_BOOL},
 	{"display_header_pane", "TRUE", &prefs_common.display_header_pane,
 	 P_BOOL},
+	{"show_attach_tab", "FALSE", &prefs_common.show_attach_tab, P_BOOL},
+	{"show_attached_files_first", "TRUE",
+	 &prefs_common.show_attached_files_first, P_BOOL},
+	{"attach_toolbutton_pos", "0", &prefs_common.attach_toolbutton_pos,
+	 P_INT},
 	{"display_header", "TRUE", &prefs_common.display_header, P_BOOL},
 	{"render_html", "TRUE", &prefs_common.render_html, P_BOOL},
 	{"html_only_as_attach", "FALSE", &prefs_common.html_only_as_attach,
@@ -357,11 +370,11 @@ static PrefParam param[] = {
 	/* Junk mail */
 	{"enable_junk", "FALSE", &prefs_common.enable_junk, P_BOOL},
 #ifdef G_OS_WIN32
-	{"junk_learn_command", "bsfilterw -C -s -u",
+	{"junk_learn_command", "sylfilter -j",
 	 &prefs_common.junk_learncmd, P_STRING},
-	{"nojunk_learn_command", "bsfilterw -c -S -u",
+	{"nojunk_learn_command", "sylfilter -c",
 	 &prefs_common.nojunk_learncmd, P_STRING},
-	{"junk_classify_command", "bsfilterw",
+	{"junk_classify_command", "sylfilter",
 	 &prefs_common.junk_classify_cmd, P_STRING},
 #else
 	{"junk_learn_command", "bogofilter -N -s -I",
@@ -493,6 +506,11 @@ static PrefParam param[] = {
 	 P_BOOL},
 	{"io_timeout_secs", "60", &prefs_common.io_timeout_secs, P_INT},
 
+	/* File selector */
+	{"filesel_prev_open_dir", NULL, &prefs_common.prev_open_dir, P_STRING},
+	{"filesel_prev_save_dir", NULL, &prefs_common.prev_save_dir, P_STRING},
+	{"filesel_prev_folder_dir", NULL, &prefs_common.prev_folder_dir, P_STRING},
+
 	{NULL, NULL, NULL, P_OTHER}
 };
 
