@@ -1,6 +1,6 @@
 /*
  * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
- * Copyright (C) 1999-2010 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2012 Hiroyuki Yamamoto
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -61,6 +61,7 @@
 #include "account.h"
 #include "account_dialog.h"
 #include "procmsg.h"
+#include "procheader.h"
 #include "filter.h"
 #include "send_message.h"
 #include "inc.h"
@@ -152,7 +153,7 @@ static struct RemoteCmd {
 			  mainwin->mainwin_cid); \
 }
 
-#ifdef G_OS_WIN32
+#if defined(G_OS_WIN32) || defined(__APPLE__)
 static void fix_font_setting		(void);
 #endif
 
@@ -175,7 +176,10 @@ static gboolean lock_socket_input_cb	(GIOChannel	*source,
 					 gpointer	 data);
 
 static void remote_command_exec		(void);
+
+#if !defined(G_OS_WIN32) && !defined(__APPLE__)
 static void migrate_old_config		(void);
+#endif
 
 static void open_compose_new		(const gchar	*address,
 					 GPtrArray	*attach_files);
@@ -278,7 +282,7 @@ int main(int argc, char *argv[])
 		 gtk_major_version, gtk_minor_version, gtk_micro_version,
 		 TARGET_ALIAS);
 
-#ifdef G_OS_WIN32
+#if defined(G_OS_WIN32) || defined(__APPLE__)
 	fix_font_setting();
 #endif
 
@@ -438,7 +442,9 @@ static void read_ini_file(void)
 		cmd.configdir = TRUE;
 	}
 }
+#endif /* G_OS_WIN32 */
 
+#if defined(G_OS_WIN32) || defined(__APPLE__)
 static void fix_font_setting(void)
 {
 	const gchar *str = NULL;
@@ -457,15 +463,24 @@ static void fix_font_setting(void)
 
 	debug_print("fixing prefs_common.textfont setting\n");
 
+#ifdef G_OS_WIN32
 	str = "MS Gothic 12";
+#else /* __APPLE__ */
+	str = "Hiragino Kaku Gothic Pro Light 13";
+#endif
 
 	if (!gtkut_font_can_load(str)) {
+#ifdef G_OS_WIN32
 		debug_print("font '%s' load failed\n", str);
 		str = "\xef\xbc\xad\xef\xbc\xb3 \xe3\x82\xb4\xe3\x82\xb7\xe3\x83\x83\xe3\x82\xaf 12";
 		if (!gtkut_font_can_load(str)) {
 			debug_print("font '%s' load failed\n", str);
 			str = NULL;
 		}
+#else /* __APPLE__ */
+		debug_print("font '%s' load failed\n", str);
+		str = NULL;
+#endif
 	}
 
 	if (str) {
@@ -552,6 +567,8 @@ static void parse_cmd_opt(int argc, char *argv[])
 			const gchar *p = argv[i + 1];
 
 			if (p && *p != '\0' && *p != '-') {
+				if (cmd.open_msg)
+					g_free(cmd.open_msg);
 				cmd.open_msg = g_locale_to_utf8
 					(p, -1, NULL, NULL, NULL);
 				i++;
@@ -595,10 +612,10 @@ static void parse_cmd_opt(int argc, char *argv[])
 		} else if (!strncmp(argv[i], "--help", 6)) {
 			init_console();
 
-			g_print(_("Usage: %s [OPTION]...\n"),
+			g_print(_("Usage: %s [OPTIONS ...] [URL]\n"),
 				g_basename(argv[0]));
 
-			g_print("%s\n", _("  --compose [address]    open composition window"));
+			g_print("%s\n", _("  --compose [mailto URL] open composition window"));
 			g_print("%s\n", _("  --attach file1 [file2]...\n"
 				"                         open composition window with specified files\n"
 				"                         attached"));
@@ -608,7 +625,8 @@ static void parse_cmd_opt(int argc, char *argv[])
 			g_print("%s\n", _("  --status [folder]...   show the total number of messages"));
 			g_print("%s\n", _("  --status-full [folder]...\n"
 				"                         show the status of each folder"));
-			g_print("%s\n", _("  --open folderid/msgnum open message in new window"));
+			g_print("%s\n", _("  --open folderid/msgnum open existing message in a new window"));
+			g_print("%s\n", _("  --open <file URL>      open an rfc822 message file in a new window"));
 			g_print("%s\n", _("  --configdir dirname    specify directory which stores configuration files"));
 #ifdef G_OS_WIN32
 			g_print("%s\n", _("  --ipcport portnum      specify port for IPC remote commands"));
@@ -626,6 +644,21 @@ static void parse_cmd_opt(int argc, char *argv[])
 
 			cleanup_console();
 			exit(1);
+		} else {
+			/* file or URL */
+			const gchar *p = argv[i];
+
+			if (p && *p != '\0') {
+				if (!strncmp(p, "mailto:", 7)) {
+					cmd.compose = TRUE;
+					cmd.compose_mailto = p + 7;
+				} else {
+					if (cmd.open_msg)
+						g_free(cmd.open_msg);
+					cmd.open_msg = g_locale_to_utf8
+						(p, -1, NULL, NULL, NULL);
+				}
+			}
 		}
 	}
 
@@ -735,7 +768,7 @@ static void parse_gtkrc_files(void)
 
 static void setup_rc_dir(void)
 {
-#ifndef G_OS_WIN32
+#if !defined(G_OS_WIN32) && !defined(__APPLE__)
 	CHDIR_EXIT_IF_FAIL(get_home_dir(), 1);
 
 	/* backup if old rc file exists */
@@ -775,7 +808,7 @@ static void setup_rc_dir(void)
 		if (is_dir_exist(OLD_RC_DIR))
 			migrate_old_config();
 	}
-#endif /* !G_OS_WIN32 */
+#endif /* !G_OS_WIN32 && !__APPLE__ */
 
 	syl_setup_rc_dir();
 }
@@ -1647,6 +1680,12 @@ static void remote_command_exec(void)
 	}
 
 	if (!gtkut_window_modal_exist()) {
+		if (cmd.compose)
+			open_compose_new(cmd.compose_mailto, cmd.attach_files);
+
+		if (cmd.open_msg)
+			open_message(cmd.open_msg);
+
 		if (cmd.receive_all)
 			inc_all_account_mail(mainwin, FALSE);
 		else if (prefs_common.chk_on_startup)
@@ -1654,14 +1693,8 @@ static void remote_command_exec(void)
 		else if (cmd.receive)
 			inc_mail(mainwin);
 
-		if (cmd.compose)
-			open_compose_new(cmd.compose_mailto, cmd.attach_files);
-
 		if (cmd.send)
 			send_queue();
-
-		if (cmd.open_msg)
-			open_message(cmd.open_msg);
 	}
 
 	if (cmd.attach_files) {
@@ -1686,6 +1719,7 @@ static void remote_command_exec(void)
 	}
 }
 
+#if !defined(G_OS_WIN32) && !defined(__APPLE__)
 static void migrate_old_config(void)
 {
 	GDir *dir;
@@ -1763,6 +1797,7 @@ static void migrate_old_config(void)
 	g_pattern_spec_free(pspec);
 	g_dir_close(dir);
 }
+#endif /* !G_OS_WIN32 && !__APPLE__ */
 
 static void open_compose_new(const gchar *address, GPtrArray *attach_files)
 {
@@ -1778,6 +1813,7 @@ static void open_compose_new(const gchar *address, GPtrArray *attach_files)
 		utf8addr = g_locale_to_utf8(address, -1, NULL, NULL, NULL);
 		if (utf8addr)
 			g_strstrip(utf8addr);
+		debug_print("open compose: %s\n", utf8addr ? utf8addr : "");
 	}
 
 #ifdef G_OS_WIN32
@@ -1806,36 +1842,87 @@ static void open_compose_new(const gchar *address, GPtrArray *attach_files)
 	g_free(utf8addr);
 }
 
+static void open_message_file(const gchar *file)
+{
+	MsgInfo *msginfo;
+	MsgFlags flags = {0};
+	MessageView *msgview;
+
+	g_return_if_fail(file != NULL);
+
+	debug_print("open message file: %s\n", file);
+
+	if (!is_file_exist(file) || get_file_size(file) <= 0) {
+		debug_print("file not found: %s\n", file);
+		return;
+	}
+
+	msginfo = procheader_parse_file(file, flags, FALSE);
+	if (msginfo) {
+		msginfo->file_path = g_strdup(file);
+		msgview = messageview_create_with_new_window();
+		messageview_show(msgview, msginfo, FALSE);
+		procmsg_msginfo_free(msginfo);
+	} else
+		debug_print("cannot open message: %s\n", file);
+}
+
 static void open_message(const gchar *path)
 {
-	gchar *id;
+	gchar *fid;
 	gchar *msg;
 	gint num;
 	FolderItem *item;
 	MsgInfo *msginfo;
 	MessageView *msgview;
+	gchar *file;
+
+	g_return_if_fail(path != NULL);
 
 	if (gtkut_window_modal_exist())
 		return;
 
-	id = g_path_get_dirname(path);
+	debug_print("open message: %s\n", path);
+
+	if (!strncmp(path, "file:", 5)) {
+		file = g_filename_from_uri(path, NULL, NULL);
+		open_message_file(file);
+		g_free(file);
+		return;
+	} else if (g_path_is_absolute(path)) {
+		open_message_file(path);
+		return;
+	}
+
+	/* relative path, or folder identifier */
+
+	fid = g_path_get_dirname(path);
 	msg = g_path_get_basename(path);
 	num = to_number(msg);
-	item = folder_find_item_from_identifier(id);
-	debug_print("open folder id: %s (msg %d)\n", id, num);
+	item = folder_find_item_from_identifier(fid);
 
 	if (num > 0 && item) {
+		debug_print("open folder id: %s (msg %d)\n", fid, num);
 		msginfo = folder_item_get_msginfo(item, num);
 		if (msginfo) {
 			msgview = messageview_create_with_new_window();
 			messageview_show(msgview, msginfo, FALSE);
 			procmsg_msginfo_free(msginfo);
+			g_free(msg);
+			g_free(fid);
+			return;
 		} else
 			debug_print("message %d not found\n", num);
 	}
 
 	g_free(msg);
-	g_free(id);
+	g_free(fid);
+
+	/* relative path */
+
+	file = g_strconcat(get_startup_dir(), G_DIR_SEPARATOR_S, path, NULL);
+	open_message_file(file);
+	g_free(file);
 }
 
 static void send_queue(void)
